#ifndef IR_NUMBER_GENERATOR_H
#define IR_NUMBER_GENERATOR_H

#include <vector>
#include "Ogre.h"
#include <ctime>
#include <cstdlib>

namespace IR
{
class NumberGenerator
{
public:
	NumberGenerator() {}
	NumberGenerator(int num) { n = num; }
	~NumberGenerator() {}

protected:
	
	int n; // Number of elements
	
};

// Divides into a uniform NxN grid 
class UniformGenerator: public NumberGenerator
{
public:
	static UniformGenerator& getInstance() 
	{
		static UniformGenerator u(100);
		return u;
	}
	~UniformGenerator() {}

	const std::vector<Ogre::Vector2>& getPairs() { return pairs; }

	void generate(int n) 
	{
		int m;
		for (m=1;m*m<=n;m++);

		for (int i=0; i<m; i++)
		{
			for (int j=0;j<m;j++)
			{
				pairs.push_back(Ogre::Vector2(i/(1.0*m),j/(1.0*m)));
			}
		}
	}

protected:
	UniformGenerator(int n): NumberGenerator(n) { generate(n); }
	UniformGenerator() {}
	std::vector<Ogre::Vector2> pairs; // Pairs of numbers
};

#define MAX_PRIMES 100

// Halton numbers
class HaltonGenerator: public NumberGenerator
{
public:	
	~HaltonGenerator() {}

	static HaltonGenerator& getInstance() 
	{
		static HaltonGenerator h(200);
		return h;
	}

	const std::vector<Ogre::Vector2> getPairs(int n, int p1, int p2)
	{
		pairs.clear();
		for (int i=0;i<n;i++)
			pairs.push_back(Ogre::Vector2(sequences[p1][i],sequences[p2][i]));

		return pairs;
	}

protected:
	HaltonGenerator(int n) 
	{		
		generate(n); 
	}

	// Returns if number is a prime
	bool isPrime(int n)
	{
		for (int i=2;i*i <= n; i++)
		{
			if (n%i == 0) return false;
		}

		return true;
	}

	// Returns the nth prime
	int getPrime(int n)
	{
		int numPrimes = 0;
		for (int i=2;;i++)
		{
			if (isPrime(i))
			{
				numPrimes++;
				if (numPrimes == n) return i;
			}			
		}
	}

	// Generates Halton sequences for MAX_PRIMES primes
	void generate(int n) 
	{
		for (int i=1;i<=MAX_PRIMES;i++)
		{
			int p = getPrime(i);
			for (int j=1;j<=n;j++)
			{
				sequences[i].push_back(getHaltonNumber(p,j));
			}
		}
	}

	// Direct calculation of Halton number [KEL99]
	double getHaltonNumber(int b, int i)
	{
		double x =0.0;
		double f = 1/ (b*1.0);

		while (i)
		{
			x += f * (double)(i%b);
			i /= b;
			f *= 1/(b*1.0);
		}
		return x;
	}

	// Incremental calcuation of Halton number
	double getHaltonNumber(int b, double x)
	{
		double h ,hh , r = 1.0 - x - 1e-10;
		double binv = 1/(b*1.0);
		if (binv < r)
		{
			x+= binv;
		}
		else
		{
			h = binv;

			do
			{
				 hh = h;
				 h *= binv;
			}while (h>=r);

			x += hh + h - 1.0;
		}
		return x;
	}
		
	std::vector<double> sequences[MAX_PRIMES];
	std::vector<Ogre::Vector2> pairs;
};


#define MAX_REFLECTIONS 25
// Divides into a uniform NxN grid 
class PoissonDiskGenerator: public NumberGenerator
{
public:
	static PoissonDiskGenerator& getInstance() 
	{		
		static PoissonDiskGenerator u(200);
		return u;
	}
	~PoissonDiskGenerator() {}

	const std::vector<Ogre::Vector2>& getPairs() { return pairs[reflection]; }

	void generate(int n) 
	{
		reflection = 0;
		srand (2007); 
		for (int count =0; count < MAX_REFLECTIONS; count++)
		{
			int precision = 1000;
			float radius = 0.008;
			do
			{
				float x = (rand()%precision)/1000.0f;
				float y = (rand()%precision)/1000.0f;				

				for (unsigned int i = 0U; i<pairs[count].size();i++)
				{
					if ( (pairs[count][i].x - x)*(pairs[count][i].x - x) + 
						(pairs[count][i].y - y)*(pairs[count][i].y - y) > radius*radius)
					{
						pairs[count].push_back(Ogre::Vector2(x,y));						
						break;
					}
				}
				
				if (pairs[count].size() == 0) pairs[count].push_back(Ogre::Vector2(x,y));
				// Is the generated number
			}while(pairs[count].size()< (unsigned int)n);
		}
	}

	void setReflection(int r)
	{
		reflection = r;
	}

protected:
	PoissonDiskGenerator(int n): NumberGenerator(n) { reflection = 0;}
	PoissonDiskGenerator() {}

	int reflection;
	std::vector<Ogre::Vector2> pairs[MAX_REFLECTIONS] ; // 20 reflections 
};

}

#endif

