#include "IRGameState.h"

#include <conio.h>
#include <iostream>
#include <string>
#include "GearApplication.h"
#include "GearInputServer.h"
#include "GearCamera.h"
#include "GearWide5Camera.h"
#include "GearSingleCamera.h"
#include "GearVideoCameraServer.h"
#include "GearTrackerServer.h"
#include "GearTimeServer.h"
#include "GearSceneLoader.h"
#include "GearSoundServer.h"
#include "GearPhysicsServer.h"
#include "IRNumberGenerator.h"
#include "IRUtility.h"

#ifndef DEGREES_TO_RADIANS
#define DEGREES_TO_RADIANS (3.141592 / 180.0)
#endif

#define ITERATIONS_PER_RATE_REPORT 300

#define MAX_RADIOSITY_LIGHTS 1000

#define BALL_MESH "ball.mesh"

#define SOME_STUPID_ATTENUATION_FACTOR_FOR_THE_STUPID_MOVABLE_LIGHT 0.65

namespace IR
{

GameState::GameState()
	: m_RayCaster(NULL), m_BackgroundImage(NULL),
	m_IRLights(NULL), m_IRWandLamp(NULL), m_IRRays(NULL),
	m_IRLightNodes(NULL), m_IRRayNodes(NULL),
	m_IRRootNode(NULL), m_IRStaticSourceLightNode(NULL), m_IRMovableSourceLightNode(NULL),
	m_IRRayMaterial(NULL),
	m_bShadowDebugging(false), m_iOnFrameIterationCounter(0),
	m_bRadiosity(false), m_bPhotonDisplay(false), m_bPrintLightValues(false),
	m_iNumSamples(-1), m_iNumPhotons(-1), m_iNumRays(-1),
	m_fRho(0.0), m_iNumLightSourceTriangles(-1), m_eSamplingType(UNIFORM),
	m_fCustomAttenationForRays(1.0), m_iRayDepthDisplayed(-1),
	m_bUseWandForLight(false) {}

GameState::~GameState(void) {}

//------------------------------------------------------------------------------
/**

*/
void GameState::onAttachToApplication() {}

//------------------------------------------------------------------------------
/**

*/
void GameState::onRemoveFromApplication() {}

//------------------------------------------------------------------------------
/**
Enter State, Load Level
*/
void GameState::onStateEnter(Ogre::String prevState)
{
	//Read config file
	Ogre::ConfigFile configFile;
	try
	{
		configFile.load(Ogre::String("..//..//resources//config//ARInstantRadiosity.cfg"));
	}
	catch (Ogre::Exception)
	{
		Ogre::LogManager::getSingleton().logMessage(Ogre::String("Error loading config file "), LML_CRITICAL);
	}

	//Create scene manager
	Ogre::SceneManager* sceneMgr = Ogre::Root::getSingletonPtr()->getSceneManager(m_sSceneManager);

	Ogre::Root::getSingletonPtr()->getRenderSystem()->setShadingType(Ogre::SO_PHONG);

	//Create camera
	Gear::Camera* cam = Gear::Application::getSingletonPtr()->getCamera();

	//Create physics server
	Gear::PhysicsServer::getSingletonPtr()->onStateEnter(prevState);

	//Calculate offsets specified in .scenario file
	Ogre::Vector3 position = Ogre::Vector3::ZERO;
	position.x = Ogre::StringConverter::parseReal(configFile.getSetting("Xpos", "Scene"));
	position.y = Ogre::StringConverter::parseReal(configFile.getSetting("Ypos", "Scene"));
	position.z = Ogre::StringConverter::parseReal(configFile.getSetting("Zpos", "Scene"));

	const Ogre::Vector3 xAxis = Ogre::Vector3::UNIT_X;
	const Ogre::Vector3 yAxis = Ogre::Vector3::UNIT_Y;
	const Ogre::Vector3 zAxis = Ogre::Vector3::UNIT_Z;
	Ogre::Real aroundXDegrees = Ogre::StringConverter::parseReal(configFile.getSetting("Xrot", "Scene"));
	Ogre::Real aroundYDegrees = Ogre::StringConverter::parseReal(configFile.getSetting("Yrot", "Scene"));
	Ogre::Real aroundZDegrees = Ogre::StringConverter::parseReal(configFile.getSetting("Zrot", "Scene"));
	const Ogre::Radian aroundXRadians(DEGREES_TO_RADIANS * aroundXDegrees);
	const Ogre::Radian aroundYRadians(DEGREES_TO_RADIANS * aroundYDegrees);
	const Ogre::Radian aroundZRadians(DEGREES_TO_RADIANS * aroundZDegrees);
	Ogre::Quaternion aroundXQuat(aroundXRadians, xAxis);
	Ogre::Quaternion aroundYQuat(aroundYRadians, yAxis);
	Ogre::Quaternion aroundZQuat(aroundZRadians, zAxis);
	Ogre::Quaternion orientation = aroundZQuat * (aroundYQuat * aroundXQuat);

	Ogre::Vector3 scale = Ogre::Vector3::UNIT_SCALE;
	Ogre::String scaleString = configFile.getSetting("Scale", "Scene");
	if(scaleString != "")
		scale.x = scale.y = scale.z = Ogre::StringConverter::parseReal(scaleString);

	//Create scene loader and load current scene
	Gear::SceneLoader* sceneLoader = new Gear::SceneLoader(sceneMgr);
	sceneLoader->setPosition(position);
	sceneLoader->setOrientation(orientation);
	sceneLoader->setScale(scale);
	Ogre::String levelName = configFile.getSetting("Name", "Scene");
	sceneLoader->parseScene(levelName, ResourceGroupManager::DEFAULT_RESOURCE_GROUP_NAME); 
	delete sceneLoader;

	Ogre::String shadowTechnique = configFile.getSetting("ShadowTechnique", "Scene");
	if (shadowTechnique == "" || shadowTechnique == "None")
		sceneMgr->setShadowTechnique(SHADOWTYPE_NONE);
	else if (shadowTechnique == "Stencil_Modulative")
		sceneMgr->setShadowTechnique(SHADOWTYPE_STENCIL_MODULATIVE);
	else if (shadowTechnique == "Stencil_Additive")
		sceneMgr->setShadowTechnique(SHADOWTYPE_STENCIL_ADDITIVE);
	else if (shadowTechnique == "Texture_Modulative")
		sceneMgr->setShadowTechnique(SHADOWTYPE_TEXTURE_MODULATIVE);
	else if (shadowTechnique == "Texture_Additive")
		sceneMgr->setShadowTechnique(SHADOWTYPE_TEXTURE_ADDITIVE);
	sceneMgr->setShowDebugShadows(m_bShadowDebugging);

	//TODO: Why call this here?
	cam->onStateEnter();
	cam->enableCompositor(false);

	//Create background image with video feed
	m_BackgroundImage = new Gear::BackgroundImage();
	m_BackgroundImage->setMaterial("Background/Video");
	m_BackgroundImage->onStateEnter();

	//Camera framerate settings and graphics synchronization
	Gear::VideoCameraServer::getSingletonPtr()->setParameters(false, 0);

	m_iOnFrameIterationCounter = 0;

	m_bRadiosity = Ogre::StringConverter::parseBool(configFile.getSetting("Radiosity", "Scene"));
	m_bPhotonDisplay = false;
	m_iNumSamples = Ogre::StringConverter::parseInt(configFile.getSetting("N", "Scene"));
	m_fRho = Ogre::StringConverter::parseReal(configFile.getSetting("Rho", "Scene"));
	m_fCustomAttenationForRays = Ogre::StringConverter::parseReal(configFile.getSetting("CustomAttenationForRays", "Scene"));

	Ogre::String sSamplingType = configFile.getSetting("SamplingType", "Scene");
	if(sSamplingType == "Poisson_Disk")
		m_eSamplingType = POISSON_DISK;
	else if(sSamplingType == "Halton_Sequence")
		m_eSamplingType = HALTON_SEQUENCE;
	else
		m_eSamplingType = UNIFORM;

	if(m_bRadiosity)
		this->instantRadiosityOnStateEnter();
	else
	{
		Ogre::Light* testLight = sceneMgr->createLight("TestLight");
		testLight->setDiffuseColour(1.0, 1.0, 1.0);
		testLight->setType(Ogre::Light::LT_POINT);
		testLight->setAttenuation(5000, 1.0, 0.0, 0.0);
		Ogre::Entity* testLightSphere = sceneMgr->createEntity("TestLightSphere", BALL_MESH);
		Ogre::SceneNode* originNode = sceneMgr->getSceneNode("CornellBoxOrigin");
		Ogre::SceneNode* testLightNode = originNode->createChildSceneNode("TestLightNode", Ogre::Vector3(-1.0, -0.5, -2.5));
		testLightNode->attachObject(testLight);
		testLightNode->attachObject(testLightSphere);
		testLightNode->setScale(0.25, 0.25, 0.25);
	}
}

//------------------------------------------------------------------------------
/**

*/
void GameState::onStateLeave(Ogre::String nextState)
{
	if(m_bRadiosity)
		this->instantRadiosityOnStateLeave();

	Gear::Application::getSingletonPtr()->getCamera()->onStateLeave();

	m_BackgroundImage->onStateLeave();
	delete m_BackgroundImage;

	Gear::PhysicsServer::getSingletonPtr()->onStateLeave(nextState);

	Ogre::SceneManager* sceneMgr = Ogre::Root::getSingleton().getSceneManager(m_sSceneManager);
	sceneMgr->destroyAllEntities();
	sceneMgr->getRootSceneNode()->removeAndDestroyAllChildren();
	sceneMgr->destroyAllLights();	
}

//------------------------------------------------------------------------------
/**
Handle User Input / steer camera
*/
Ogre::String GameState::onFrame()
{
	Gear::Application *application = Gear::Application::getSingletonPtr();

	//---VIDEO---
	Gear::VideoCameraServer *videoCameraServer = Gear::VideoCameraServer::getSingletonPtr();

	if(videoCameraServer->onFrame() == Gear::FRAMEEVENT_NEW_FRAME)
	{
		videoCameraServer->updateVideoFrame();
		videoCameraServer->saveVideoFrame();
	}


	//---TRACKING---
	if (application->hasTrackingSupport())
	{
		Gear::TrackerServer *trackerServer = Gear::TrackerServer::getSingletonPtr();
		trackerServer->updateTrackerData();
		trackerServer->saveTrackerData();
		trackerServer->onFrame();
	}


	//---INPUT---
	Gear::InputServer *inputServer = Gear::InputServer::getSingletonPtr();
	inputServer->onFrame();

	//---WAND BUTTONS---
	if(inputServer->isWandButtonPressed(0) || inputServer->isWandButtonPressed(1))
		m_bUseWandForLight = !m_bUseWandForLight;


	//---KEYBOARD---
	if(inputServer->isKeyTriggered("Quit"))
		return "Exit";
	else if(inputServer->isKeyTriggered("ToggleShadowDebugging"))
		this->toggleShadowDebugging();
	else if(inputServer->isKeyTriggered("ToggleTracking"))
		Gear::Application::getSingletonPtr()->getCamera()->toggleTracking();
	else if(inputServer->isKeyTriggered("TooglePhotonDisplay"))
		m_bPhotonDisplay = !m_bPhotonDisplay;
	else if(inputServer->isKeyTriggered("ToggleCoordCross"))
		Gear::Application::getSingletonPtr()->getCamera()->showDebugCoordCross(true);
	else if(inputServer->isKeyTriggered("ShowAllLights"))
		m_iRayDepthDisplayed = -1;
	else if(inputServer->isKeyTriggered("ShowSourceLightsOnly"))
		m_iRayDepthDisplayed = 0;
	else if(inputServer->isKeyTriggered("ShowFirstBounceLightsOnly"))
		m_iRayDepthDisplayed = 1;
	else if(inputServer->isKeyTriggered("ShowSecondBounceLightsOnly"))
		m_iRayDepthDisplayed = 2;
	else if(inputServer->isKeyTriggered("ShowThirdBounceLightsOnly"))
		m_iRayDepthDisplayed = 3;
	else if(inputServer->isKeyTriggered("ShowFourthBounceLightsOnly"))
		m_iRayDepthDisplayed = 4;
	else if(inputServer->isKeyTriggered("ShowFifthBounceLightsOnly"))
		m_iRayDepthDisplayed = 5;
	else if(inputServer->isKeyTriggered("ShowSixthBounceLightsOnly"))
		m_iRayDepthDisplayed = 6;
	else if(inputServer->isKeyTriggered("ShowSeventhBounceLightsOnly"))
		m_iRayDepthDisplayed = 7;
	else if(inputServer->isKeyTriggered("ShowEighthBounceLightsOnly"))
		m_iRayDepthDisplayed = 8;
	else if(inputServer->isKeyTriggered("ShowNinthBounceLightsOnly"))
		m_iRayDepthDisplayed = 9;
	else if(inputServer->isKeyTriggered("PrintLightValues"))
		m_bPrintLightValues = true;


	//---CAMERA CONTROL---
	Gear::Application::getSingletonPtr()->getCamera()->onFrame();


	//---HUD---
	application->getDebugHUD()->updateStats();

	if(m_bRadiosity)
		this->instantRadiosityOnFrame();

	//Render the scene (this does per-light passes and additive light masking by itself
	application->getOgreRoot()->renderOneFrame();


	//---RATE REPORT---
	if((++m_iOnFrameIterationCounter % ITERATIONS_PER_RATE_REPORT) == 0)
		_cprintf("-- CAMERA: %0.1f native, %0.1f effective --\n", videoCameraServer->getNativeFramerate(), videoCameraServer->getEffectiveFramerate());


	return m_StateName;
}

void GameState::toggleShadowDebugging(void)
{
	m_bShadowDebugging = !m_bShadowDebugging;
	Ogre::Root::getSingletonPtr()->getSceneManager(m_sSceneManager)->setShowDebugShadows(m_bShadowDebugging);
}

void GameState::instantRadiosityOnStateEnter(void)
{
	//Compute the total number of lights (after all bounces) and rays
	m_iNumPhotons = MAX_RADIOSITY_LIGHTS;
	m_iNumRays = MAX_RADIOSITY_LIGHTS;

	// Modify the exported materials from Blender whose names start with "IRMat_" to have additive light masking
	Ogre::MaterialManager::ResourceMapIterator materialResourceMapIterator = Ogre::MaterialManager::getSingleton().getResourceIterator();
	while (materialResourceMapIterator.hasMoreElements())
	{
		Ogre::MaterialPtr materialPtr = materialResourceMapIterator.getNext();
		Ogre::String materialName = materialPtr->getName();

		if(materialName.find("IRMat_", 0) != std::string::npos)
		{
			Ogre::Technique* techniquePtr = materialPtr->getTechnique(0);
			Ogre::Pass* ambientPassPtr = techniquePtr->getPass(0);

			//get diffuse color (as exported from Blender
			Ogre::ColourValue diffuseColor = ambientPassPtr->getDiffuse();

			//turn the pass exported by Blender into very dim ambient light pass
			ambientPassPtr->setDiffuse(0.0, 0.0, 0.0, 1.0);
			ambientPassPtr->setSpecular(0.0, 0.0, 0.0, 1.0);

			//add a diffuse pass per light
			Ogre::Pass* diffusePassPtr = techniquePtr->createPass();
			diffusePassPtr->setMaxSimultaneousLights(m_iNumPhotons);
			diffusePassPtr->setIteratePerLight(true);
			diffusePassPtr->setSceneBlending(Ogre::SBT_ADD);
			diffusePassPtr->setDepthWriteEnabled(true);
			diffusePassPtr->setDiffuse(diffuseColor);
			//diffusePassPtr->setAmbient(0.1, 0.1, 0.1);
			diffusePassPtr->setSpecular(Ogre::ColourValue::Black);
			diffusePassPtr->setLightingEnabled(true);
		}
	}

	//Create material for the rays and wand lamp
	m_IRRayMaterial = Ogre::MaterialManager::getSingleton().create("rayMaterial","debugger");
	m_IRRayMaterial->setReceiveShadows(false);
	m_IRRayMaterial->getTechnique(0)->setLightingEnabled(true);
	m_IRRayMaterial->getTechnique(0)->getPass(0)->setDiffuse(0,0,1,0);
	m_IRRayMaterial->getTechnique(0)->getPass(0)->setAmbient(0,0,1);
	m_IRRayMaterial->getTechnique(0)->getPass(0)->setSelfIllumination(0,0,1);

	//Create nodes for the lights
	Ogre::SceneManager* sceneMgr = Ogre::Root::getSingletonPtr()->getSceneManager(m_sSceneManager);
	Ogre::SceneNode* originNode = (Ogre::SceneNode*)(sceneMgr->getRootSceneNode()->getChild("CornellBoxOrigin"));
	this->m_IROffsetFromWorldCoords = originNode->getWorldPosition();

	m_IRRootNode = originNode->createChildSceneNode("RadiosityRootNode");
	m_IRLightNodes = new Ogre::SceneNode*[m_iNumPhotons];

	//Create the lights
	m_IRLights = new Ogre::Light*[m_iNumPhotons];
	for(int k = 0; k < m_iNumPhotons; k++)
	{
		m_IRLights[k] = sceneMgr->createLight(Ogre::String("RadiosityLight") + Ogre::StringConverter::toString(k));
		m_IRLights[k]->setType(Ogre::Light::LT_POINT);
		m_IRLights[k]->setAttenuation(5000, 1.0, 0.0, 0.0);
		m_IRLights[k]->setCastShadows(true);
		m_IRLights[k]->setVisible(false);
		m_IRLightNodes[k] = m_IRRootNode->createChildSceneNode(Ogre::String("RadiosityLightNode") + Ogre::StringConverter::toString(k));
		m_IRLightNodes[k]->attachObject(m_IRLights[k]);

	}

	//Build light source samples
	IR::PoissonDiskGenerator::getInstance().generate(500);

	this->m_IRStaticSourceLightNode = sceneMgr->getSceneNode("IRLight_area light");
	this->m_IRSourceLightOrigPosition = m_IRStaticSourceLightNode->getPosition();
	this->m_IRSourceLightOrigOrientation = m_IRStaticSourceLightNode->getOrientation();
	m_iNumLightSourceTriangles = 0;

	for (int j=0; j < m_IRStaticSourceLightNode->numAttachedObjects(); j++)
	{
		MovableObject* mObject = m_IRStaticSourceLightNode->getAttachedObject(j);

		String s = mObject->getMovableType();
		if (mObject->getMovableType().compare("Entity")==0)
		{
			// get the entity to check
			Entity* pEntity = static_cast<Ogre::Entity*>(mObject);           

			// mesh data to retrieve         
			size_t vertex_count;
			size_t index_count;
			Ogre::Vector3 *vertices;
			unsigned long *indices;

			// get the mesh information
			GetMeshInformation(pEntity->getMesh(), vertex_count, vertices, index_count, indices,
				Ogre::Vector3::ZERO, Ogre::Quaternion::IDENTITY, Ogre::Vector3::UNIT_SCALE); // This is computed still in Blender's coord system, apply offset quat later

			// Now sample each face of the light 
			for (unsigned int k = 0U; k<index_count; k+=3)
			{
				_cprintf("%f %f %f\n",vertices[indices[k]].x,vertices[indices[k]].y, vertices[indices[k]].z);

				m_LightSourceTriangles[m_iNumLightSourceTriangles].m_Vertices[0] = vertices[indices[k]]; 
				m_LightSourceTriangles[m_iNumLightSourceTriangles].m_Vertices[1] = vertices[indices[k+1]];
				m_LightSourceTriangles[m_iNumLightSourceTriangles].m_Vertices[2] = vertices[indices[k+2]];

				m_LightSourceTriangles[m_iNumLightSourceTriangles].m_Normal =
					Ogre::Math::calculateBasicFaceNormal(vertices[indices[k]], vertices[indices[k+1]], vertices[indices[k+2]]);

				m_LightSourceTriangles[m_iNumLightSourceTriangles].m_Tangent = vertices[indices[k]] - vertices[indices[k+1]];
				m_LightSourceTriangles[m_iNumLightSourceTriangles].m_Tangent.normalise();

				Ogre::String sLightSourceMaterialName = pEntity->getMesh()->getSubMesh(0)->getMaterialName();
				Ogre::MaterialPtr pLightSourceMaterial =
					static_cast<Ogre::MaterialPtr>(Ogre::MaterialManager::getSingletonPtr()->getByName(sLightSourceMaterialName));
				m_LightSourceTriangles[m_iNumLightSourceTriangles].m_Color = pLightSourceMaterial->getTechnique(0)->getPass(0)->getDiffuse();

				m_LightSourceTriangles[m_iNumLightSourceTriangles].computeSamples(m_iNumSamples / 2, m_eSamplingType);   // <- HACK!
				m_iNumLightSourceTriangles++;
			}
		}
	}

	//Create rays and respective scene nodes
	m_IRRays = new Ogre::ManualObject*[m_iNumRays];
	m_IRRayNodes = new Ogre::SceneNode*[m_iNumRays];
	for(int k = 0; k < m_iNumRays; k++)
	{
		m_IRRays[k] = sceneMgr->createManualObject(Ogre::String("Ray") + Ogre::StringConverter::toString(k));
		m_IRRayNodes[k] = m_IRRootNode->createChildSceneNode(Ogre::String("RayNode") + Ogre::StringConverter::toString(k));
		m_IRRayNodes[k]->attachObject(m_IRRays[k]);
	}

	//Create wand lamp and respective node
	m_IRWandLamp = sceneMgr->createEntity("WandLamp", BALL_MESH);
	m_IRWandLamp->setMaterialName("rayMaterial");
	m_IRWandLamp->setCastShadows(false);
	m_IRWandLamp->setVisible(false);
	m_IRMovableSourceLightNode = m_IRRootNode->createChildSceneNode("MovableLightSource");
	m_IRMovableSourceLightNode->attachObject(m_IRWandLamp);
	m_IRMovableSourceLightNode->setScale(0.25, 0.25, 0.25);

	//Create raycaster object
	m_RayCaster = new RayCaster();
	m_RayCaster->onStateEnter(m_sSceneManager);
}

void GameState::instantRadiosityOnStateLeave(void)
{
	Ogre::SceneManager* sceneMgr = Ogre::Root::getSingletonPtr()->getSceneManager(m_sSceneManager);
	for(int k = 0; k < m_iNumRays; k++)
	{
		m_IRRayNodes[k]->detachObject(m_IRRays[k]);
		m_IRRootNode->removeAndDestroyChild(Ogre::String("RayNode") + Ogre::StringConverter::toString(k));
		sceneMgr->destroyEntity(Ogre::String("Ray") + Ogre::StringConverter::toString(k));
	}

	m_RayCaster->onStateLeave();
	delete m_RayCaster;
}

void GameState::instantRadiosityOnFrame(void)
{
	if(m_bRadiosity)
	{
		int lightCount = 0;
		int rayCount = 0;

		Ogre::SceneManager* sceneMgr = Ogre::Root::getSingletonPtr()->getSceneManager(m_sSceneManager);

		for(int k = 0; k < m_iNumLightSourceTriangles; k++)
		{
			int N = m_LightSourceTriangles[k].numSamples;

			//Algorithm straight out of the paper
			Ogre::Real w, start;
			int end, reflections = 0;
			Ogre::ColourValue L;
			Ogre::Vector3 y, omega;

			start = end = N;

			while(end > 0)
			{
				start *= m_fRho;

				for(int i = (int)start; i < end; i++)
				{
					Ogre::Quaternion lightSourceOrientation;
					Ogre::Vector3 lightSourcePosition;
					Ogre::Vector3 y, normal, tangent;
					Ogre::ColourValue L;

					if(m_bUseWandForLight)
					{
						//We're using the wand to update the node's position and orientation
						Gear::TrackerServer* trackerServer = Gear::TrackerServer::getSingletonPtr();
						Ogre::Quaternion wandOrientation = trackerServer->getViewMatrix(1).extractQuaternion(); // in world coordinates
						Ogre::Vector3 wandPosition = trackerServer->getViewMatrix(1).getTrans(); // in world coordinates
						
						lightSourceOrientation = wandOrientation * m_IRSourceLightOrigOrientation;
						lightSourcePosition = wandPosition - m_IROffsetFromWorldCoords;

						y = lightSourceOrientation * (m_LightSourceTriangles[k].m_Samples[i] * m_IRMovableSourceLightNode->getScale()) + lightSourcePosition;
						normal = m_IRStaticSourceLightNode->getOrientation() * m_LightSourceTriangles[k].m_Normal;
						tangent = m_IRStaticSourceLightNode->getOrientation() * m_LightSourceTriangles[k].m_Tangent;
						L = SOME_STUPID_ATTENUATION_FACTOR_FOR_THE_STUPID_MOVABLE_LIGHT  * m_LightSourceTriangles[0].m_Color;

						m_IRMovableSourceLightNode->setOrientation(lightSourceOrientation);
						m_IRMovableSourceLightNode->setPosition(lightSourcePosition);
						m_IRWandLamp->setVisible(true);
					}
					else
					{
						//We're using the room's ceiling light
						lightSourceOrientation = m_IRSourceLightOrigOrientation;
						lightSourcePosition = m_IRSourceLightOrigPosition;
						m_IRWandLamp->setVisible(false);

						y = lightSourceOrientation * (m_LightSourceTriangles[k].m_Samples[i] * m_IRStaticSourceLightNode->getScale()) + lightSourcePosition;
						normal = m_IRStaticSourceLightNode->getOrientation() * m_LightSourceTriangles[k].m_Normal;
						tangent = m_IRStaticSourceLightNode->getOrientation() * m_LightSourceTriangles[k].m_Tangent;
						L = m_LightSourceTriangles[k].m_Color;
					}

					w = (Ogre::Real)N * m_iNumLightSourceTriangles;

					//trace reflections
					for(int j = 0; j <= reflections; j++)
					{
						if(m_iRayDepthDisplayed == -1 || m_iRayDepthDisplayed == j)
						{
							//Place node in correct position
							m_IRLightNodes[lightCount]->setPosition(y);

							//Set light parameters
							Ogre::ColourValue lightColor = L / floorf(w);
							m_IRLights[lightCount]->setDiffuseColour(lightColor);
							m_IRLights[lightCount]->setSpecularColour(Ogre::ColourValue::Black);
							m_IRLights[lightCount]->setVisible(true);
							m_IRLights[lightCount]->setCastShadows(j == 0); //Only first pass is used for shadowing

							if(m_bPrintLightValues)
							{
								_cprintf("Light %d:\t(%0.4f, %0.4f, %0.4f)\n", lightCount, lightColor.r, lightColor.g, lightColor.b);
							}
						}
						else
							m_IRLights[lightCount]->setVisible(false);

						lightCount++;

						if(reflections > 0 && j < reflections)
						{
							std::vector<Vector2> results;
							
							switch(m_eSamplingType)
							{
							case UNIFORM:
								results = IR::UniformGenerator::getInstance().getPairs();
								break;
							case POISSON_DISK:
								IR::PoissonDiskGenerator::getInstance().setReflection(j);
								results = IR::PoissonDiskGenerator::getInstance().getPairs();
								break;
							case HALTON_SEQUENCE:
								//results = IR::HaltonGenerator::getInstance().getPairs(m_LightSourceTriangles[k].numSamples, 2 * j + 2, 2 * j + 3);
								results = IR::HaltonGenerator::getInstance().getPairs(m_LightSourceTriangles[k].numSamples, j + 2, j + 3);
								break;
							default:
								throw;
							}

							Ogre::Ray ray;
							m_RayCaster->createRay(&ray, y + m_IROffsetFromWorldCoords, normal, tangent, results[i].x, results[i].y, !m_bUseWandForLight);

							Ogre::Vector3 intersectionPosition, intersectionNormal, intersectionTangent;
							Ogre::Entity* intersectionEntity = NULL;

							m_IRRays[rayCount]->setVisible(false);
							m_IRRays[rayCount]->clear();

							if (m_RayCaster->intersectRay(ray, &intersectionPosition, &intersectionNormal, &intersectionTangent, &intersectionEntity))
							{
								if(m_bPhotonDisplay && (m_iRayDepthDisplayed == -1 || m_iRayDepthDisplayed == j + 1))
								{
									m_IRRays[rayCount]->begin(m_IRRayMaterial->getName(), Ogre::RenderOperation::OT_LINE_LIST);
										m_IRRays[rayCount]->position(y);
										m_IRRays[rayCount]->position(intersectionPosition - m_IROffsetFromWorldCoords);
									m_IRRays[rayCount]->end();
									m_IRRays[rayCount]->setVisible(true);
									rayCount++;
								}	

								y = intersectionPosition - m_IROffsetFromWorldCoords;
								normal = intersectionNormal;
								tangent = intersectionTangent;

								Ogre::String intersectionMaterialName = intersectionEntity->getMesh()->getSubMesh(0)->getMaterialName();
								Ogre::MaterialPtr intersectionMaterial =
									static_cast<Ogre::MaterialPtr>(Ogre::MaterialManager::getSingletonPtr()->getByName(intersectionMaterialName));
								Ogre::Technique* intersectionMaterialTechnique = intersectionMaterial->getTechnique(0);
								Ogre::ColourValue intersectionColor;
								if(intersectionMaterialTechnique->getNumPasses() == 2)
									intersectionColor = intersectionMaterialTechnique->getPass(1)->getDiffuse();
								else
									intersectionColor = intersectionMaterialTechnique->getPass(0)->getDiffuse();

								L = L * intersectionColor * m_fCustomAttenationForRays; // custom hack

								w *= m_fRho;
							}
						}
					}

				}
				reflections++;		
				end = (int)start;
			}
		}

		for(int k = rayCount; k < m_iNumRays; k++)
		{
			m_IRRays[k]->setVisible(false);
			m_IRRays[k]->clear();
		}

		m_bPrintLightValues = false;
	}
}

}; // end namespace