#include "IRRayCaster.h"

#include "OgreRoot.h"
#include "OgrePrerequisites.h"
#include "OgreEntity.h"
#include "OgreSubMesh.h"
#include "IRUtility.h"

// For some reason this isn't working, and it should!
//#define _USE_MATH_DEFINES
//#include <cmath>

#define PI 3.14159265358979323846

namespace IR {

RayCaster::RayCaster() : m_pSceneManager(NULL), m_pRaySceneQuery(NULL) {}

void RayCaster::onFrame() {}

void RayCaster::onStateEnter(Ogre::String sceneManager)
{		
	m_pSceneManager = Ogre::Root::getSingletonPtr()->getSceneManager(sceneManager);
	m_pRaySceneQuery = NULL;
	m_sceneOrigin = m_pSceneManager->getRootSceneNode()->getChild("CornellBoxOrigin")->getPosition();
	m_sceneOrientation = m_pSceneManager->getRootSceneNode()->getChild("CornellBoxOrigin")->getOrientation();
}

void RayCaster::onStateLeave(void) {}

Ogre::Vector3 RayCaster::generateDirection(const Ogre::Real phi2, const Ogre::Real phi3, bool isHemisphere) const
{
	Ogre::Real zenith;
	
	if (isHemisphere)
		zenith = asin(sqrt(phi2));
	else
		zenith = acos(1 - 2 * phi2);

	Ogre::Real azimuth = phi3 * 2 * PI;

	//I use the tangent as my x axis
	Ogre::Vector3 projectedDirection(cos(azimuth), sin(azimuth), 0.0);

	//I use the normal as my z axis
	Ogre::Real cosZenith = cos(zenith);
	Ogre::Real x = cosZenith * projectedDirection.x;
	Ogre::Real y = cosZenith * projectedDirection.y;
	Ogre::Real z = sin(zenith);
	Ogre::Vector3 direction(x, y, z);
	
	//just in case
	return direction;
}

bool RayCaster::createRay(Ogre::Ray* ray,
						  const Ogre::Vector3& origin,
						  const Ogre::Vector3& normal,
						  const Ogre::Vector3& tangent,
						  const Ogre::Real phi2,
						  const Ogre::Real phi3,
						  bool isHemisphere) const
{
	Ogre::Vector3 directionOnLocalCoords = this->generateDirection(phi2, phi3, isHemisphere);

	Ogre::Vector3 binormal = normal.crossProduct(tangent);
	Ogre::Matrix3 worldCoordSystem;
	worldCoordSystem.FromAxes(tangent, binormal, normal);

	Ogre::Vector3 directionOnGlobalCoords = worldCoordSystem * directionOnLocalCoords;
	
	ray->setOrigin(origin);
	ray->setDirection(directionOnGlobalCoords);
	return true;
}

bool RayCaster::intersectRay(const Ogre::Ray& ray,
							 Ogre::Vector3* intersection,
							 Ogre::Vector3* normal,
							 Ogre::Vector3* tangent,
							 Ogre::Entity** entity)
{
	if(m_pRaySceneQuery != NULL)
		m_pSceneManager->destroyQuery(m_pRaySceneQuery);
	m_pRaySceneQuery = m_pSceneManager->createRayQuery(ray, Ogre::SceneManager::WORLD_GEOMETRY_TYPE_MASK);
	if (m_pRaySceneQuery == NULL)
		throw;
	m_pRaySceneQuery->setSortByDistance(true); 

	// check we are initialised
	if (m_pRaySceneQuery != NULL)
	{
		// create a query object
		m_pRaySceneQuery->setRay(ray);

		int numResults = (int)((m_pRaySceneQuery->execute()).size());
		// execute the query, returns a vector of hits
		if (numResults <= 0)
		{
			// raycast did not hit an objects bounding box
			return (false);
		}
	}
	else
	{
		//LOG_ERROR << "Cannot raycast without RaySceneQuery instance" << ENDLOG;
		return (false);
	}   

	// at this point we have raycast to a series of different objects bounding boxes.
	// we need to test these different objects to see which is the first polygon hit.
	// there are some minor optimizations (distance based) that mean we wont have to
	// check all of the objects most of the time, but the worst case scenario is that
	// we need to test every triangle of every object.
	Ogre::Real closestDistance = -1.0f;
	Ogre::Vector3 closestResult;
	Ogre::Vector3 closestTriangle[3];
	Ogre::RaySceneQueryResult &queryResult = m_pRaySceneQuery->getLastResults();
	Ogre::Entity *pEntity;

	for (size_t qr_idx = 0; qr_idx < queryResult.size(); qr_idx++)
	{
		// stop checking if we have found a raycast hit that is closer
		// than all remaining entities
		if ((closestDistance >= 0.0f) &&
			(closestDistance < queryResult[qr_idx].distance))
		{
			break;
		}

		// only check this result if its a hit against an entity
		if ((queryResult[qr_idx].movable != NULL) &&
			(queryResult[qr_idx].movable->getMovableType().compare("Entity") == 0))
		{
			// get the entity to check
			pEntity = static_cast<Ogre::Entity*>(queryResult[qr_idx].movable);

			if(pEntity->getMesh()->getSubMesh(0)->getMaterialName().find("IRMat_", 0) == std::string::npos)
			{
				pEntity = NULL;
				continue;
			}

			// mesh data to retrieve         
			size_t vertex_count;
			size_t index_count;
			Ogre::Vector3 *vertices;
			unsigned long *indices;

			// get the mesh information
			GetMeshInformation(pEntity->getMesh(), vertex_count, vertices, index_count, indices,             
				pEntity->getParentNode()->getWorldPosition(),
				pEntity->getParentNode()->getWorldOrientation(),
				pEntity->getParentNode()->getScale());

			// test for hitting individual triangles on the mesh
			bool newClosestFound = false;
			Ogre::Vector3 newClosestTriangle[3];
			for (int i = 0; i < static_cast<int>(index_count); i += 3)
			{
				// check for a hit against this triangle
				std::pair<bool, Ogre::Real> hit = Ogre::Math::intersects(ray, vertices[indices[i]],
					vertices[indices[i+1]], vertices[indices[i+2]], true, false);

				// if it was a hit check if its the closest
				if (hit.first)
				{
					if ((closestDistance < 0.0f) ||
						(hit.second < closestDistance))
					{
						// this is the closest so far, save it off
						closestDistance = hit.second;
						newClosestFound = true;
						newClosestTriangle[0] = vertices[indices[i]];
						newClosestTriangle[1] = vertices[indices[i+1]];
						newClosestTriangle[2] = vertices[indices[i+2]];
					}
				}
			}

			// free the verticies and indicies memory
			delete[] vertices;
			delete[] indices;

			// if we found a new closest raycast for this object, update the
			// closestResult before moving on to the next object.
			if (newClosestFound)
			{
				closestResult = ray.getPoint(closestDistance);
				closestTriangle[0] = newClosestTriangle[0];
				closestTriangle[1] = newClosestTriangle[1];
				closestTriangle[2] = newClosestTriangle[2];
			}
		}       
	}

	// return the result
	if (closestDistance >= 0.0f && pEntity)
	{
		// raycast success
		*intersection = closestResult;

		*normal = Ogre::Math::calculateBasicFaceNormal(closestTriangle[0], closestTriangle[1], closestTriangle[2]);

		*tangent = closestTriangle[0] - closestResult;
		tangent->normalise();

		*entity = pEntity;
		return (true);
	}
	else
	{
		// raycast failed
		return (false);
	}
}

} // end namespace